from random import Random

from .mypy_types import OptInt


class Sampler:

    def __init__(self, *, sample_rate: float=1.0, seed: OptInt=None) -> None:
        self._sample_rate = sample_rate
        self._rng = Random(seed)

    def is_sampled(self, trace_id: str) -> bool:
        if self._sample_rate == 0.0:
            sampled = False
        else:
            sampled = self._rng.random() <= self._sample_rate
        return sampled

# TODO: implement other types of sampler for example hash trace_id



import time
from typing import NamedTuple, Optional, Dict, List, Any

from .mypy_types import Headers, OptBool, OptInt, OptStr, OptTs


# possible span kinds
CLIENT = 'CLIENT'
SERVER = 'SERVER'
PRODUCER = 'PRODUCER'
CONSUMER = 'CONSUMER'

# zipkin headers, for more information see:
# https://github.com/openzipkin/b3-propagation

TRACE_ID_HEADER = 'X-B3-TraceId'
SPAN_ID_HEADER = 'X-B3-SpanId'
PARENT_ID_HEADER = 'X-B3-ParentSpanId'
FLAGS_HEADER = 'X-B3-Flags'
SAMPLED_ID_HEADER = 'X-B3-Sampled'


_TraceContext = NamedTuple(
    'TraceContext', [
        ('trace_id', str),
        ('parent_id', OptStr),
        ('span_id', str),
        ('sampled', bool),
        ('debug', bool),
        ('shared', bool),
        ]
)


class TraceContext(_TraceContext):
    """Immutable class with trace related data that travels across
    process boundaries.
    """

    def make_headers(self) -> Headers:
        """Creates dict with zipkin headers from available context.

        Resulting dict should be passed to HTTP client  propagate contest
        to other services.
        """
        return make_headers(self)


Endpoint = NamedTuple(
    'Endpoint', [('serviceName', OptStr),
                 ('ipv4', OptStr),
                 ('ipv6', OptStr),
                 ('port', OptInt)]
)


def create_endpoint(servce_name: str, *,
                    ipv4: OptStr=None,
                    ipv6: OptStr=None,
                    port: OptInt=None):
    """Factory function to create Endpoint object.
    """
    return Endpoint(servce_name, ipv4, ipv6, port)


def make_timestamp(ts: OptTs=None) -> int:
    """Create zipkin timestamp in microseconds, or convert available one
    from second. Useful when user supply ts from time.time() call.
    """
    ts = ts if ts is not None else time.time()
    return int(ts * 1000 * 1000)  # microseconds


def make_headers(context: TraceContext) -> Headers:
    """Creates dict with zipkin headers from supplied trace context.
    """
    headers = {
        TRACE_ID_HEADER: context.trace_id,
        SPAN_ID_HEADER: context.span_id,
        FLAGS_HEADER: '0',
        SAMPLED_ID_HEADER: '1' if context.sampled else '0',
    }
    if context.parent_id is not None:
        headers[PARENT_ID_HEADER] = context.parent_id
    return headers


def parse_sampled(headers: Headers) -> OptBool:
    sampled = headers.get(SAMPLED_ID_HEADER.lower(), None)
    if sampled is None or sampled == '':
        return None
    return True if sampled == '1' else False


def parse_debug(headers: Headers) -> bool:
    return True if headers.get(FLAGS_HEADER, '0') == '1' else False


def make_context(headers: Headers) -> Optional[TraceContext]:
    """Converts available headers to TraceContext, if headers mapping does
    not contain zipkin headers, function returns None.
    """
    # TODO: add validation for trace_id/span_id/parent_id

    # normalize header names just in case someone passed regular dict
    # instead dict with case insensitive keys
    headers = {k.lower(): v for k, v in headers.items()}

    if not all(h in headers for h in (
            TRACE_ID_HEADER.lower(), SPAN_ID_HEADER.lower())):
        return None

    context = TraceContext(
        trace_id=headers.get(TRACE_ID_HEADER.lower()),
        parent_id=headers.get(PARENT_ID_HEADER.lower(), None),
        span_id=headers.get(SPAN_ID_HEADER.lower()),
        sampled=parse_sampled(headers),
        shared=False,
        debug=parse_debug(headers),
    )
    return context


OptKeys = Optional[List[str]]


def filter_none(data: Dict[str, Any],
                keys: OptKeys=None) -> Dict[str, Any]:
    """Filter keys from dict with None values.

    Check occurs only on root level. If list of keys specified, filter
    works only for selected keys
    """

    def limited_filter(k: str, v: Any) -> bool:
        return k not in keys or v is not None  # type: ignore

    def full_filter(k: str, v: Any) -> bool:
        return v is not None

    f = limited_filter if keys is not None else full_filter
    return {k: v for k, v in data.items() if f(k, v)}



import asyncio
from typing import List, Dict, Any, Optional  # flake8: noqa

import aiohttp
from yarl import URL

from .log import logger
from .mypy_types import OptLoop
from .record import Record


class Transport:

    def __init__(self, address: str, send_inteval: float=5,
                 loop: OptLoop=None) -> None:
        self._address = URL(address).with_path('/api/v2/spans')
        self._queue = []  # type: List[Dict[str, Any]]
        self._closing = False
        self._send_interval = send_inteval
        self._loop = loop or asyncio.get_event_loop()
        self._session = aiohttp.ClientSession(loop=self._loop)
        self._sender_task = asyncio.ensure_future(
            self._sender_loop(), loop=self._loop)

        self._ender = self._loop.create_future()
        self._timer = None  # type: Optional[asyncio.Future[Any]]

    def send(self, record: Record) -> None:
        data = record.asdict()
        self._queue.append(data)

    async def _sender_loop(self) -> None:
        while not self._ender.done():
            if self._queue:
                await self._send()

            await self._wait()

    async def _wait(self) -> None:
        self._timer = asyncio.ensure_future(
            asyncio.sleep(self._send_interval, loop=self._loop),
            loop=self._loop)

        await asyncio.wait([self._timer, self._ender], loop=self._loop,
                           return_when=asyncio.FIRST_COMPLETED)

    async def _send(self) -> None:
        # TODO: add retries
        data = self._queue[:]
        self._queue = []

        try:
            headers = {'Content-Type': 'application/json'}
            async with self._session.post(self._address, json=data,
                                          headers=headers) as resp:
                body = await resp.text()
                if resp.status >= 300:
                    msg = 'zipkin responded with code: {} and body: {}'.format(
                        resp.status, body)
                    raise RuntimeError(msg)

        except Exception as exc:
            # that code should never fail and break application
            logger.error('Can not send spans to zipkin', exc_info=exc)

    async def close(self) -> None:
        if self._closing:
            return

        self._closing = True
        self._ender.set_result(None)

        await self._sender_task
        await self._send()
        await self._session.close()

        if self._timer is not None:
            self._timer.cancel()
            try:
                await self._timer
            except asyncio.CancelledError:
                pass



from typing import TypeVar, Dict, Any, List, NamedTuple
from .mypy_types import OptInt, OptStr, Optional  # flake8: noqa

from .helpers import (
    CONSUMER,
    PRODUCER,
    Endpoint,
    TraceContext,
    filter_none,
)


Annotation = NamedTuple('Annotation', [('value', str), ('timestamp', int)])


def _endpoint_asdict(endpoint: Endpoint) -> Dict[str, Any]:
    return filter_none(endpoint._asdict())


T = TypeVar('T', bound='Record')


class Record:

    def __init__(self: T, context: TraceContext,
                 local_endpoint: Endpoint) -> None:
        self._context = context
        self._local_endpoint = _endpoint_asdict(local_endpoint)
        self._finished = False

        self._name = 'unknown'
        self._kind = None  # type: OptStr
        self._timestamp = None  # type: OptInt
        self._duration = None  # type: OptInt
        self._remote_endpoint = None  # type: Optional[Dict[str, Any]]
        self._annotations = []  # type: List[Annotation]
        self._tags = {}  # type: Dict[str, str]

    def start(self: T, ts: int) -> T:
        self._timestamp = ts
        return self

    def finish(self: T, ts: int) -> T:
        if self._finished:
            return self
        if ts is not None and self._kind not in (PRODUCER, CONSUMER):
            self._duration = max(ts - self._timestamp, 1)
        self._finished = True
        return self

    def name(self: T, n: str) -> T:
        self._name = n
        return self

    def set_tag(self: T, key: str, value: Any) -> T:
        self._tags[key] = str(value)
        return self

    def annotate(self: T, value: str, ts: int) -> T:
        self._annotations.append(Annotation(str(value), int(ts)))
        return self

    def kind(self: T, kind: str) -> T:
        self._kind = kind
        return self

    def remote_endpoint(self: T, endpoint: Endpoint) -> T:
        self._remote_endpoint = _endpoint_asdict(endpoint)
        return self

    def asdict(self) -> Dict[str, Any]:
        c = self._context
        rec = {
            'traceId': c.trace_id,
            'name': self._name,
            'parentId': c.parent_id,
            'id': c.span_id,
            'kind': self._kind,
            'timestamp': self._timestamp,
            'duration': self._duration,
            'debug': c.debug,
            'shared': c.shared,
            'localEndpoint': self._local_endpoint,
            'remoteEndpoint': self._remote_endpoint,
            'annotations': [a._asdict() for a in self._annotations],
            'tags': self._tags,
        }
        return filter_none(rec, ['kind'])



import ipaddress
from typing import Optional, Dict, Any, Set  # flake8: noqa

import aiohttp
from aiohttp.web import HTTPException, Request, Application
from aiohttp.web_urldispatcher import AbstractRoute

from .constants import HTTP_METHOD, HTTP_PATH, HTTP_STATUS_CODE
from .helpers import CLIENT, SERVER, make_context, parse_debug, parse_sampled
from .span import SpanAbc
from .tracer import Tracer


APP_AIOZIPKIN_KEY = 'aiozipkin_tracer'
REQUEST_AIOZIPKIN_KEY = 'aiozipkin_span'


__all__ = (
    'setup',
    'get_tracer',
    'request_span',
    'middleware_maker',
    'make_trace_config',
    'APP_AIOZIPKIN_KEY',
    'REQUEST_AIOZIPKIN_KEY',
)


def _set_remote_endpoint(span: SpanAbc, request: Request) -> None:
    peername = request.remote
    if peername is not None:
        kwargs = {}  # type: Dict[str, Any]
        try:
            peer_ipaddress = ipaddress.ip_address(peername)
        except ValueError:
            pass
        else:
            if isinstance(peer_ipaddress, ipaddress.IPv4Address):
                kwargs['ipv4'] = str(peer_ipaddress)
            else:
                kwargs['ipv6'] = str(peer_ipaddress)
        if kwargs:
            span.remote_endpoint(None, **kwargs)


def _get_span(request: Request, tracer: Tracer) -> SpanAbc:
    # builds span from incoming request, if no context found, create
    # new span
    context = make_context(request.headers)

    if context is None:
        sampled = parse_sampled(request.headers)
        debug = parse_debug(request.headers)
        span = tracer.new_trace(sampled=sampled, debug=debug)
    else:
        span = tracer.join_span(context)
    return span


def _set_span_properties(span: SpanAbc, request: Request) -> None:
    span_name = '{0} {1}'.format(request.method.upper(),
                                 request.path)
    span.name(span_name)
    span.kind(SERVER)
    span.tag(HTTP_PATH, request.path)
    span.tag(HTTP_METHOD, request.method.upper())
    _set_remote_endpoint(span, request)


# TODO: new aiohttp 3.0.0 has a bit different API for middlewares
# should be reworked once 3.0.0 out, do we care about backward compatibility
def middleware_maker(skip_routes: Optional[AbstractRoute]=None,
                     tracer_key=APP_AIOZIPKIN_KEY,
                     request_key=REQUEST_AIOZIPKIN_KEY):
    s = skip_routes
    skip_routes_set = set(s) if s else set()  # type: Set[AbstractRoute]

    async def middleware_factory(app, handler):
        async def aiozipkin_middleware(request):
            # route is in skip list, we do not track anything with zipkin
            if request.match_info.route in skip_routes_set:
                resp = await handler(request)
                return resp

            tracer = request.app[tracer_key]
            span = _get_span(request, tracer)
            request[request_key] = span
            if span.is_noop:
                resp = await handler(request)
                return resp

            with span:
                _set_span_properties(span, request)
                try:
                    resp = await handler(request)
                except HTTPException as e:
                    span.tag(HTTP_STATUS_CODE, e.status)
                    raise

                span.tag(HTTP_STATUS_CODE, resp.status)
            return resp

        return aiozipkin_middleware

    return middleware_factory


def setup(app: Application,
          tracer: Tracer, *,
          skip_routes: Optional[AbstractRoute]=None,
          tracer_key: str=APP_AIOZIPKIN_KEY,
          request_key: str=REQUEST_AIOZIPKIN_KEY) -> Application:
    """Sets required parameters in aiohttp applications for aiozipkin.

    Tracer added into application context and cleaned after application
    shutdown. You can provide custom tracer_key, if default name is not
    suitable.
    """
    app[tracer_key] = tracer
    m = middleware_maker(skip_routes=skip_routes,
                         tracer_key=tracer_key,
                         request_key=request_key)
    app.middlewares.append(m)

    # register cleanup signal to close zipkin transport connections
    async def close_aiozipkin(app):
        await app[tracer_key].close()

    app.on_cleanup.append(close_aiozipkin)

    return app


def get_tracer(app: Application, tracer_key: str=APP_AIOZIPKIN_KEY) -> Tracer:
    """Returns tracer object from application context.

    By default tracer has APP_AIOZIPKIN_KEY in aiohttp application context,
    you can provide own key, if for some reason default one is not suitable.
    """
    return app[tracer_key]


def request_span(request: Request,
                 request_key: str=REQUEST_AIOZIPKIN_KEY) -> SpanAbc:
    """Return span created by middleware from request context, you can use it
    as parent on next child span.
    """
    return request[request_key]


class ZipkinClientSignals:
    """Class contains signal handler for aiohttp client. Handlers executed
    only if aiohttp session contains tracer context with span.
    """

    def __init__(self, tracer: Tracer) -> None:
        self._tracer = tracer

    def _has_span(self, trace_config_ctx) -> bool:
        trace_request_ctx = trace_config_ctx.trace_request_ctx
        has_span = (isinstance(trace_request_ctx, dict) and
                    'span_context' in trace_request_ctx)
        return has_span

    async def on_request_start(self, session, context, params) -> None:
        if not self._has_span(context):
            return

        p = params
        span_context = context.trace_request_ctx['span_context']
        span = self._tracer.new_child(span_context)
        context._span = span
        span.start()
        span_name = 'client {0} {1}'.format(p.method.upper(), p.url.path)
        span.name(span_name)
        span.kind(CLIENT)

        propagate_headers = (context
                             .trace_request_ctx
                             .get('propagate_headers', True))
        if propagate_headers:
            span_headers = span.context.make_headers()
            p.headers.update(span_headers)

    async def on_request_end(self, session, context, params) -> None:
        if not self._has_span(context):
            return

        span = context._span
        span.finish()
        delattr(context, '_span')

    async def on_request_exception(self, session, context, params) -> None:
        if not self._has_span(context):
            return
        span = context._span
        span.finish(exception=params.exception)
        delattr(context, '_span')


def make_trace_config(tracer: Tracer) -> aiohttp.TraceConfig:
    trace_config = aiohttp.TraceConfig()
    zipkin = ZipkinClientSignals(tracer)

    trace_config.on_request_start.append(zipkin.on_request_start)
    trace_config.on_request_end.append(zipkin.on_request_end)
    trace_config.on_request_exception.append(zipkin.on_request_exception)
    return trace_config



from typing import Optional, Dict, Awaitable, Any, AsyncContextManager  # flake8: noqa

from .context_managers import _ContextManager
from .helpers import TraceContext, Endpoint
from .mypy_types import OptLoop, OptBool
from .record import Record
from .sampler import Sampler
from .span import Span, NoopSpan, SpanAbc
from .transport import Transport
from .utils import generate_random_64bit_string, generate_random_128bit_string


class Tracer(AsyncContextManager):
    def __init__(self,
                 transport: Transport,
                 sampler: Sampler,
                 local_endpoint: Endpoint) -> None:
        self._records = {}  # type: Dict[TraceContext, Record]
        self._transport = transport
        self._sampler = sampler
        self._local_endpoint = local_endpoint

    def new_trace(self,
                  sampled: OptBool=None,
                  debug: bool=False) -> SpanAbc:
        context = self._next_context(None, sampled=sampled, debug=debug)
        return self.to_span(context)

    def join_span(self, context: TraceContext) -> SpanAbc:
        new_context = context
        if context.sampled is None:
            sampled = self._sampler.is_sampled(context.trace_id)
            new_context = new_context._replace(sampled=sampled)
        else:
            new_context = new_context._replace(shared=True)
        return self.to_span(new_context)

    def new_child(self, context: TraceContext) -> SpanAbc:
        new_context = self._next_context(context)
        if not context.sampled:
            return NoopSpan(self, new_context)
        return self.to_span(new_context)

    def to_span(self, context: TraceContext) -> SpanAbc:
        if not context.sampled:
            return NoopSpan(self, context)

        record = Record(context, self._local_endpoint)
        self._records[context] = record
        return Span(self, context, record)

    def _send(self, record: Record) -> None:
        self._records.pop(record._context, None)
        self._transport.send(record)

    def _next_context(self,
                      context: Optional[TraceContext]=None,
                      sampled: OptBool=None,
                      debug: bool=False) ->TraceContext:
        span_id = generate_random_64bit_string()
        if context is not None:
            new_context = context._replace(
                span_id=span_id,
                parent_id=context.span_id,
                shared=False)
            return new_context

        trace_id = generate_random_128bit_string()
        if sampled is None:
            sampled = self._sampler.is_sampled(trace_id)

        new_context = TraceContext(
            trace_id=trace_id,
            parent_id=None,
            span_id=span_id,
            sampled=sampled,
            debug=debug,
            shared=False)
        return new_context

    async def close(self) -> None:
        await self._transport.close()

    async def __aenter__(self) -> 'Tracer':
        return self

    async def __aexit__(self, *args) -> Optional[bool]:
        return await self.close()


def create(zipkin_address: str,
           local_endpoint: Endpoint, *,
           sample_rate: float=0.01,
           send_inteval: float=5,
           loop: OptLoop=None) -> Awaitable[Tracer]:

    async def f() -> Tracer:
        sampler = Sampler(sample_rate=sample_rate)
        transport = Transport(zipkin_address, send_inteval=send_inteval, loop=loop)
        return Tracer(transport, sampler, local_endpoint)
    result = _ContextManager(f())  # type: Awaitable[Tracer]
    return result



from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, TypeVar, Optional

from .constants import ERROR
from .helpers import Endpoint, make_timestamp, TraceContext
from .mypy_types import OptInt, OptStr, OptTs
from .record import Record


if TYPE_CHECKING:
    from .tracer import Tracer  # flake8: noqa


T = TypeVar('T', bound='SpanAbc')


class SpanAbc(metaclass=ABCMeta):

    @property
    @abstractmethod
    def is_noop(self: T) -> bool:
        return True  # pragma: no cover

    @property
    @abstractmethod
    def context(self: T) -> TraceContext:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def tracer(self: T) -> 'Tracer':
        pass  # pragma: no cover

    @abstractmethod
    def start(self: T, ts: OptTs=None) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def finish(self: T, ts: OptTs=None,
               exception: Optional[Exception]=None) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def remote_endpoint(self: T,
                        servce_name: OptStr, *,
                        ipv4: OptStr=None,
                        ipv6: OptStr=None,
                        port: OptInt=None) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def tag(self: T, key: str, value: str) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def annotate(self: T,
                 value: str,
                 ts: OptTs=None) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def kind(self: T, span_kind: str) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def name(self: T, span_name: str) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def new_child(self: T, name: OptStr = None, kind: OptStr = None) -> T:
        pass  # pragma: no cover

    def __enter__(self: T) -> T:
        self.start()
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        self.finish(exception=exception_value)


class NoopSpan(SpanAbc):
    def __init__(self, tracer: 'Tracer', context: TraceContext) -> None:
        self._context = context
        self._tracer = tracer

    @property
    def is_noop(self) -> bool:
        return True

    @property
    def context(self) -> TraceContext:
        return self._context

    @property
    def tracer(self) -> 'Tracer':
        return self._tracer

    def start(self, ts: OptTs=None) -> 'NoopSpan':
        return self

    def finish(self, ts: OptTs=None,
               exception: Optional[Exception]=None) -> 'NoopSpan':
        return self

    def remote_endpoint(self,
                        servce_name: OptStr, *,
                        ipv4: OptStr=None,
                        ipv6: OptStr=None,
                        port: OptInt=None) -> 'NoopSpan':
        return self

    def tag(self, key: str, value: str) -> 'NoopSpan':
        return self

    def annotate(self, value: str, ts: OptTs=None) -> 'NoopSpan':
        return self

    def kind(self, span_kind: str) -> 'NoopSpan':
        return self

    def name(self, span_name: str) -> 'NoopSpan':
        return self

    def new_child(self,
                  name: OptStr = None,
                  kind: OptStr = None) -> 'NoopSpan':
        context = self._tracer._next_context(self.context)
        return NoopSpan(self.tracer, context)


class Span(SpanAbc):
    def __init__(self, tracer: 'Tracer',
                 context: TraceContext,
                 record: Record) -> None:
        self._context = context
        self._tracer = tracer
        self._record = record

    @property
    def is_noop(self) -> bool:
        return False

    @property
    def context(self) -> TraceContext:
        return self._context

    @property
    def tracer(self) -> 'Tracer':
        return self._tracer

    def start(self, ts=None) -> 'Span':
        ts = make_timestamp(ts)
        self._record.start(ts)
        return self

    def finish(self, ts: OptTs=None,
               exception: Optional[Exception]=None) -> 'Span':
        if exception is not None:
            self.tag(ERROR, str(exception))
        ts = make_timestamp(ts)
        self._record.finish(ts)
        self._tracer._send(self._record)
        return self

    def remote_endpoint(self,
                        servce_name: OptStr, *,
                        ipv4: OptStr=None,
                        ipv6: OptStr=None,
                        port: OptInt=None) -> 'Span':
        endpoint = Endpoint(servce_name, ipv4, ipv6, port)
        self._record.remote_endpoint(endpoint)
        return self

    def tag(self, key: str, value: str) -> 'Span':
        self._record.set_tag(key, value)
        return self

    def annotate(self, value: str, ts: OptTs=None) -> 'Span':
        ts = make_timestamp(ts)
        self._record.annotate(value, ts)
        return self

    def kind(self, span_kind: str) -> 'Span':
        self._record.kind(span_kind)
        return self

    def name(self, span_name: str) -> 'Span':
        self._record.name(span_name)
        return self

    def new_child(self, name: OptStr = None, kind: OptStr = None) -> 'Span':
        span = self.tracer.new_child(self.context)
        if name is not None:
            span.name(name)
        if kind is not None:
            span.kind(kind)
        return span  # type: ignore



from .aiohttp_helpers import (
    get_tracer,
    make_context,
    make_trace_config,
    middleware_maker,
    request_span,
    setup,
    APP_AIOZIPKIN_KEY,
    REQUEST_AIOZIPKIN_KEY,
)
from .helpers import (
    CLIENT,
    CONSUMER,
    PRODUCER,
    SERVER,
    create_endpoint,
)
from .sampler import Sampler
from .tracer import Tracer, create


__version__ = '0.2.0'
__all__ = (
    'Tracer',
    'Sampler',
    'create',
    'create_endpoint',
    'make_context',
    # aiohttp helpers
    'setup',
    'get_tracer',
    'request_span',
    'middleware_maker',
    'make_trace_config',
    'APP_AIOZIPKIN_KEY',
    'REQUEST_AIOZIPKIN_KEY',
    # possible span kinds
    'CLIENT',
    'SERVER',
    'PRODUCER',
    'CONSUMER',
)



# standard tags (binary annotations)
HTTP_HOST = 'http.host'
HTTP_METHOD = 'http.method'
HTTP_PATH = 'http.path'
HTTP_REQUEST_SIZE = 'http.request.size'
HTTP_RESPONSE_SIZE = 'http.response.size'
HTTP_STATUS_CODE = 'http.status_code'
HTTP_URL = 'http.url'

ERROR = 'error'
LOCAL_COMPONENT = 'lc'

CLIENT_ADDR = 'ca'
MESSAGE_ADDR = 'ma'
SERVER_ADDR = 'sa'



from collections.abc import Awaitable
from typing import (TypeVar, Optional, Generator, Any, Type,
                    AsyncContextManager, Generic)
from types import TracebackType


T = TypeVar('T', bound=AsyncContextManager)


class _ContextManager(Generic[T], AsyncContextManager, Awaitable):

    __slots__ = ('_coro', '_obj')

    def __init__(self, coro: Awaitable) -> None:
        self._coro = coro  # type: Awaitable[T]
        self._obj = None  # type: Optional[T]

    def __await__(self) -> Generator[Any, None, T]:
        return self._coro.__await__()

    async def __aenter__(self) -> T:
        self._obj = await self._coro
        return await self._obj.__aenter__()

    async def __aexit__(self,
                        exc_type: Optional[Type[BaseException]],
                        exc: Optional[BaseException],
                        tb: Optional[TracebackType]) -> Optional[bool]:
        if self._obj is None:
            raise RuntimeError('__aexit__ called before __aenter__')
        return await self._obj.__aexit__(exc_type, exc, tb)
